use clap::{command, Parser};
use clap_stdin::FileOrStdin;
use std::{
    fs::{self, File},
    io::{self, BufRead, BufReader, Read, Seek, SeekFrom},
    os::linux::fs::MetadataExt,
    path::Path,
    thread, usize,
};

fn modf(value: f64) -> (f64, f64) {
    (value.trunc(), value.fract())
}

fn count_words(input: &mut impl BufRead) -> io::Result<usize> {
    let mut was_ws = true;
    let mut count = 0usize;

    loop {
        let buf = input.fill_buf()?;
        let len = buf.len();
        if len == 0 {
            break;
        }

        for &b in buf {
            let is_ws = matches!(b, b' ' | b'\t' | b'\r' | b'\n');
            count += usize::from(!was_ws && is_ws);
            was_ws = is_ws;
        }

        input.consume(len);
    }

    count += usize::from(!was_ws);
    Ok(count)
}

fn count_file<P: AsRef<Path>>(path: P) -> io::Result<usize> {
    fs::read(path).and_then(|v| count_words(&mut v.as_slice()))
}

fn count_file2<P: AsRef<Path> + std::marker::Send + Copy>(
    path: P,
    threads: usize,
) -> io::Result<usize> {
    let md = fs::metadata(path)?;
    let blksize = md.st_blksize() as usize;
    let len: usize = md.len().try_into().expect("could not read file length");
    let div = ((len / threads) as f64).ceil() as usize;

    let scope = thread::scope(|scope| {
        let mut handles = Vec::with_capacity(threads);

        for i in 0..threads {
            let handle = scope.spawn(move || {
                let mut file = File::open(&path).expect("unable to open file");
                let mut buf = vec![0u8; blksize];
                let mut read_len = 1usize;
                let mut total = 0usize;

                let offset = (i * div) as u64;

                file.seek(SeekFrom::Start(offset))
                    .expect("could not seek to position");

                while total < div && read_len != 0 {
                    if total + blksize > div {
                        buf.truncate(div - total);
                    }

                    read_len = file.read(&mut buf).expect("could not read file");
                    total += read_len;
                }

                count_words(&mut BufReader::new(buf.as_slice())).expect("could not read input")
            });

            handles.push(handle);
        }

        let mut count = 0usize;
        for handle in handles {
            count += handle.join().expect("failed to read from thread handle");
        }

        count
    });

    Ok(scope)
}

#[derive(Parser)]
#[command(version, about, long_about = None)]
struct Cli {
    #[arg(short, long, default_value_t = 200)]
    speed: usize,
    #[arg(short, long, default_value_t = num_cpus::get())]
    threads: usize,
    input: FileOrStdin,
}

fn main() {
    let cli = Cli::parse();

    let count = if cli.input.is_file() {
        //count_file(&cli.input.filename()).expect("could not read file")
        count_file2(&cli.input.filename(), cli.threads).expect("could not read file")
    } else {
        let mut buf = cli
            .input
            .into_reader()
            .map(BufReader::new)
            .expect("could not read from stdin");

        count_words(&mut buf).expect("could not read from stdin")
    };
    let time = count as f64 / cli.speed as f64;
    let (mins, fract) = modf(time);
    let secs = (fract * 0.60f64 * 100f64).ceil();

    println!("{count} words {mins}m{secs}s");
}
